/* -----------------------------------------------------------------------
   Digitemp Class code written by Anonymous

   Copyright 1997 by Brian C. Lane

   Thanks to Anonymous, who is a far better C programmer than I
   =======================================================================
   01/16/99     Fixed a 1-off error in the calc_temp routine. This caused
                it to not work below 0C!

   08/21/97     Adding comments to code, cleaning up. I think that class
                organization can be seperated somewhat so that the
                classes don't have to be so dependant on one another. Also
                could facilitate rewrite to include DOS and Linux C++
                versions all in one code tree.


   ----------------------------------------------------------------------- */
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <windows.h>

#define NO_TEMP (-1000.f)

/* Define a class to handle all of the comm port operations                */
class CommPort
{
  public:
    HANDLE  cm;                                 // Comm Port Handle
    
    inline int isopen() {return cm != INVALID_HANDLE_VALUE;}

    CommPort(int port = 0, unsigned long baud = 57600)
    {
        cm = INVALID_HANDLE_VALUE;              //start clean
        if (port != 0)
            open(port, baud);
    }

    void close()
    {
        if (!isopen())
            return;
        CloseHandle(cm);
        cm = INVALID_HANDLE_VALUE;
    }

    ~CommPort()
    {
        close();
    }

    int setcomm(unsigned long baud = 57600,
                int bits = 8, char parity='N', int stop = 1,
                long readtimeoutms = 100, long writetimeoutms = 4000)
    {
        char settings[128];
        sprintf(settings, "baud=%lu data=%d parity=%c stop=%d", 
            baud,
            bits,
            parity,
            stop);

        DCB dcb;
        if (!GetCommState(cm, &dcb) || !BuildCommDCB(settings, &dcb))
        {
            return 0;
        }

        dcb.fOutX = dcb.fInX = 0;
        dcb.fNull = 0;
        dcb.fBinary = 1;
        dcb.fDtrControl = DTR_CONTROL_ENABLE;

        if (!SetCommState(cm, &dcb))
        {
            return 0;
        }

        COMMTIMEOUTS ct;
        ct.ReadIntervalTimeout = MAXDWORD;
        ct.ReadTotalTimeoutMultiplier = MAXDWORD;
        ct.ReadTotalTimeoutConstant = readtimeoutms;

        ct.WriteTotalTimeoutMultiplier = 0;
        ct.WriteTotalTimeoutConstant = writetimeoutms;
        if (!SetCommTimeouts(cm, &ct))
        {
            return 0;
        }
        return 1;
    }

    int open(int port, unsigned long baud = 57600,
                int bits = 8, char parity='N', int stop = 1,
                long readtimeoutms = 250, long writetimeoutms = 4000)
    {
        close();
        char fname[32];
        sprintf(fname, "COM%d", port);
        cm = CreateFile(
            fname,                      //LPCTSTR lpFileName,                // pointer to name of the file 
            GENERIC_READ|GENERIC_WRITE, //    DWORD dwDesiredAccess,         // access (read-write) mode 
            0,                          //DWORD dwShareMode,                 // share mode 
            NULL,                       //LPSECURITY_ATTRIBUTES lpSecurityAttributes, // pointer to security descriptor 
            OPEN_EXISTING,              //DWORD dwCreationDistribution,      // how to create 
            FILE_ATTRIBUTE_NORMAL,      //DWORD dwFlagsAndAttributes,        // file attributes 
            NULL);                      // handle to file with attributes to copy  

        if (cm == INVALID_HANDLE_VALUE)
        {
            return 0;
        }

        if (!setcomm(baud, bits, parity, stop, readtimeoutms, writetimeoutms))
        {
            close();
            return 0;
        }
        return 1;
    }

    long write(char *text, long len)
    {
        DWORD writtencount;
        BOOL res = WriteFile(
            cm,                         //HANDLE hFile,                   // handle to file to write to 
            text,                       // lpBuffer,                      // pointer to data to write to file 
            (DWORD)len,                 //DWORD nNumberOfBytesToWrite,    // number of bytes to write 
            &writtencount,              //LPDWORD lpNumberOfBytesWritten, // pointer to number of bytes written 
            NULL);                      //LPOVERLAPPED lpOverlapped       // pointer to structure needed for overlapped I/O
        return res ? (long)writtencount : -1L;
    }

    long write(char *text)
    {
        return write(text, strlen(text));
    }
    
    long read(char *text, long len)
    {
        DWORD readcount;
        BOOL res = ReadFile(
            cm,                         //HANDLE hFile,                   // handle to file to write to 
            text,                       // lpBuffer,                      // pointer to data to write to file 
            (DWORD)len,                 //DWORD nNumberOfBytesToWrite,    // number of bytes to write 
            &readcount,                 //LPDWORD lpNumberOfBytesWritten, // pointer to number of bytes written 
            NULL);                      //LPOVERLAPPED lpOverlapped       // pointer to structure needed for overlapped I/O
        return res ? (long)readcount : -1L;
    }
};


class DS1820;                           // Predefine because it is referenced next

/* Create a derived class for DS1820 comm port activities                  */
class CommPortDS1820 : public CommPort {
public:
    DS1820  *list;                      // List of attached sensors

    int TouchReset()
    {
        // Reset the attached devices by pulling data low for > 450mS
        // The Dallas Examples use 10473 but 9600 works just as well (longer pulse)
        if (!setcomm(
            9600,
            8, 'N', 1, 3000, 3000))
            return -1;

        unsigned char byte = 0xf0;
        write((char*)&byte, 1);
        int ret = read((char*)&byte, 1);

        // Return the speed to 115200 for normal communications
        if (!setcomm(
            115200, 
            8, 'N', 1, 1500, 3000))
            return -1;
        return  -(ret != 1);
    }

    //caveat: we return first bit received in 1<<0, second in 1<<1..
    //no matter how many bits are asked for.
    int TouchBits(unsigned char outch, int bits = 8)
    {
        //we write a *byte* so the device can see a *bit*
        unsigned char mask = 01;
        unsigned char inv = 0;
        for (int i = 0; i < bits; ++i)
        {
            unsigned char bytev = (outch & mask) ? 0xff : 0x00;
            if (write((char*)&bytev, 1) != 1)
                return -1;
            if (read((char*)&bytev, 1) != 1)
                return -2;
            inv |= (bytev & 0x01) ? mask : 0;
            mask <<= 1;
        }
        return inv;
    }

    // Read/Write a whole byte to the OneWire bus
    inline int TouchByte(unsigned char outch)
    {
        return TouchBits(outch);
    }

    // Search for attached devices
    int find_thermometers();

    // Initalization: cp = comport 1..4
    CommPortDS1820(int cp) : CommPort()
    {
        open(cp, 115200, 8, 'N', 1, 1500);
        list = NULL;
        find_thermometers();
    }
};

/* Define a class that holds the data for each attached device */
class DS1820
{
  friend class CommPortDS1820;
  public:
    float               last_tempC;             // Last temperature read
    DS1820              *next;                  // Pointer to the next in the list
    char                *myname;                // Pointer to a name
    unsigned char       id[8];                  // Serial Number
    unsigned char       scratchpad[9];          // scratchpad data

protected:
    CommPortDS1820      *comport;               // Comm port it is attached to

public:

    static inline float ctof(float temp)
    {
        return 32.f + (temp * 9.f) / 5.f;
    }

    static inline float roundhalf(float temp, float to = 2.f)
    {
        int t = (int)(temp * to);
        return t / to;
    }

    // Talk to the device directly
    int MatchROM()
    {
        if (comport->TouchReset() < 0)                  // Reset all devices
            return -1;

        if (comport->TouchByte(0x55) < 0)               //match Rom command
            return -2;

        //send serial number id
        for (int x = 0; x < 8; ++x)
            if (comport->TouchByte(id[x]) < 0)
                return -3;

        return 0;
    }

    // Read the temperature from this device
    int ReadTemperature()
    {
        MatchROM();

        if (comport->TouchByte(0x44) < 0)               //calculate command
            return -1;

        Sleep(550);                                     //spec says 500 is enough

        if (comport->TouchReset() < 0)                  // Reset all devices
            return -2;

        return 0;
    }

    int ReadScratchpad();

    DS1820(CommPortDS1820 *cp, unsigned char *idp)
    {
        comport = cp;
        memcpy(id, idp, sizeof id);
        scratchpad[7] = 0;
        last_tempC = NO_TEMP;
        myname = NULL;
        next = NULL;
    }

    void setname(char *p)
    {
        free(myname);
        myname = strdup(p);
    }

    ~DS1820()
    {
        free(myname);
        myname = NULL;
    }


/* -----------------------------------------------------------------------
   Return the high-precision temperature value

   Calculated using formula from DS1820 datasheet

   Temperature   = scratchpad[0]
   Sign          = scratchpad[1]
   TH            = scratchpad[2]
   TL            = scratchpad[3]
   Count Remain  = scratchpad[6]
   Count Per C   = scratchpad[7]
   CRC           = scratchpad[8]
   
                   count_per_C - count_remain
   (temp - 0.25) * --------------------------
                       count_per_C

   If Sign is not 0x00 then it is a negative (Centigrade) number, and
   the temperature must be subtracted from 0x100 and multiplied by -1
      
   ----------------------------------------------------------------------- */
    float calc_temp()
    {
        float temp;

        if (scratchpad[1] == 0)
            temp = (float)(int)(scratchpad[0] >> 1);
        else
            temp = (float)-(int)((0x100 - scratchpad[0]) >> 1);

        if (scratchpad[7] != 0)
        {            
            temp -= 0.25f;
            temp += (float)((int)scratchpad[7] - (int)scratchpad[6]) / (float)(int)scratchpad[7];
        }
        return temp;
    }

    // Start a temperature conversion and read the scratchpad
    float tempC()
    {
        if (ReadTemperature() < 0)
            return NO_TEMP;                     //-1000.f
        if (ReadScratchpad() < 0)
            return NO_TEMP;                     //-1000.f
        last_tempC = calc_temp();
        return last_tempC;
    }  
};


/* ------------------------------------------------------------------------
   CRC table for Dallas 8-bit CRC calculations from App. Note #27
------------------------------------------------------------------------ */
static const unsigned char crc_table[] = {
     0, 94,188,226, 97, 63,221,131,194,156,126, 32,163,253, 31, 65,
   157,195, 33,127,252,162, 64, 30, 95,  1,227,189, 62, 96,130,220,
    35,125,159,193, 66, 28,254,160,225,191, 93,  3,128,222, 60, 98,
   190,224,  2, 92,223,129, 99, 61,124, 34,192,158, 29, 67,161,255,
    70, 24,250,164, 39,121,155,197,132,218, 56,102,229,187, 89,  7,
   219,133,103, 57,186,228,  6, 88, 25, 71,165,251,120, 38,196,154,
   101, 59,217,135,  4, 90,184,230,167,249, 27, 69,198,152,122, 36,
   248,166, 68, 26,153,199, 37,123, 58,100,134,216, 91,  5,231,185,
   140,210, 48,110,237,179, 81, 15, 78, 16,242,172, 47,113,147,205,
    17, 79,173,243,112, 46,204,146,211,141,111, 49,178,236, 14, 80,
   175,241, 19, 77,206,144,114, 44,109, 51,209,143, 12, 82,176,238,
    50,108,142,208, 83, 13,239,177,240,174, 76, 18,145,207, 45,115,
   202,148,118, 40,171,245, 23, 73,  8, 86,180,234,105, 55,213,139,
    87,  9,235,181, 54,104,138,212,149,203, 41,119,244,170, 72, 22,
   233,183, 85, 11,136,214, 52,106, 43,117,151,201, 74, 20,246,168,
   116, 42,200,150, 21, 75,169,247,182,232, 10, 84,215,137,107, 53,
};


/* As we understand it...
* A Search ROM command (0xF)) puts all the DS1820s into
* a mode where you can ask them to respond, in chorus,
* with the value they have for each of 64 bit positions of their ids.
* For each bit, they announce, together, whether 
* they all have a 1 in that position, or all have a 0,
* or conflict (some 1's, some 0's).
* If they all agree, just play that bit back and continue.
* If they disagree, pick an untried value (1 first)
* and play that bit back, and continue.
***
* The units that see a bit value played back
* they don't agree with fall silent at that point, allowing
* each pass through the bits to narrow down to a single
* DS1820 each time. When there's a conflict in a position,
* try the other bit value next time, and treat all bit 
* positions to the right of it as unknown, since we are
* walking a new part of the tree.
* The last bit queried always comes back with a definitive
* answer, since there's only one DS1820 selected at that
* point, and it has to say yes or no.
*/
int CommPortDS1820::find_thermometers()
{
    unsigned char unknowns[64];
    int count = 0; //we don't have any yet.

    // Initalize unknown bit markers for all 64 bits
    memset(unknowns, 0xff, sizeof unknowns);

    for (unsigned int done = 0; !done; )
    {
        if (TouchReset() < 0)
            return -1;
        if (TouchByte(0xF0) < 0)        /* ROM Search Command           */
            return -2;

        unsigned char rom_buf[8];

        //Clear out the ROM buffer
        memset(rom_buf, 0, sizeof(rom_buf));

        register int x;

        // Find the serial number for one ROM
        for (x = 0; x < 64; ++x)
        {
            //OK, DS1820's, are you all 1 or 0 in the next position?
            //Those with a 1 play back 01, thouse with a 0 play
            //back 10, and the effect is AND'd, so we see 0 if
            //there's a disagreement. We write 2 bits so they
            //can play back 2 bits.
            unsigned char bits = (unsigned char)TouchBits(0xFF, 2);
            switch (bits)
            {
            case 00:
                /* The devices don't agree about this position.
                * Send and record what we haven't tried yet (1 first),
                * If that's a 1, we'll be back here at some point
                * to try zero.
                */
                bits = unknowns[x];
                // & with 1 to get the proposed value in this position
                break;

            case 01: // All have a 1 bit in this posiiton AKA 1
            case 02: // All have a 0 bit in this position AKA 2
                // & with 1 to get the actual value in this position
                unknowns[x] = 0;        // Known data point
                break;

            default:  //either nothing's attached (3) or TouchBits is fried
                return -1;
            } /* switch */

            /* Play back the bit value we chose (the low order
            * bit in 'bits' has it here.)
            * Here, any DS1820s that don't agree with what
            * we send will fall silent, allowing us to query
            * the rest.
            */
            TouchBits(bits, 1);

            // And record our choice
            rom_buf[x / 8] |= (bits & 1) << (x % 8);
        }

        // Check the checksum on the ROM # found and move to next if ok
        unsigned char crc = 0;
        for (x = 0; x < 8; ++x)
            crc = crc_table[crc ^ rom_buf[x]];

        if (crc == 0) //all OK?
        {
            //got one! Add it to the list
            DS1820 *th = new DS1820(this, rom_buf);
            char buf[8 * 2 + 1];
            sprintf(buf, "%02x%02x%02x%02x%02x%02x%02x%02x",
                th->id[0], th->id[1], th->id[2], th->id[3],
                th->id[4], th->id[5], th->id[6], th->id[7]);
            //set the name to its id for now.
            th->setname(buf);
            th->next = list;
            list = th;
            count++;
        }
        else
        {
            printf("CRC Error (%02X)", crc);
            printf(" -- Skipping it\n");
            continue;
        }

        done = 1; // Done until proven not

        // Fill with T's up to rightmost T, and clear that one.
        //look familiar? It's a 2's complement decrement op,
        //for 64 bits, done longhand. But here it's just a
        //treewalk
        //if we come in with all F's, we leave with all T's,
        // but done remains set in the case, so we are finished.
        for (x = sizeof(unknowns) - 1; done && x >= 0; --x)
        {
            unknowns[x] = ~unknowns[x]; //T->F, F->T
            done &= unknowns[x]; //and stop if it came in as T
        }
    }

    if (TouchReset() < 0)
        return -1;

    return count;
}


int DS1820::ReadScratchpad()
{
    MatchROM();

    if (comport->TouchByte(0xBE) < 0)           //read scratchpad
        return -1;

    unsigned char crc = 0;
    for (int x = 0; x < sizeof(scratchpad); ++x)
    {
        int byte = comport->TouchByte(0xFF);
        if (byte < 0)
            return -2;
        scratchpad[x] = (unsigned char)byte;
        crc = crc_table[crc ^ scratchpad[x]];
    }

    comport->TouchReset();
    if (crc != 0)
        return -3;
    return 0;
}

